from Akatosh import Mundus, Actor

class mom(Actor):

    def __init__(self, at, step, till, priority, name) -> None:
        super().__init__(at=at, step=step, till=till, priority=priority)
        self._name = name
        self._num_children = 0

    def action(self):
        c = child(at=self.timeline.now, step=1, till=8, priority=-self._num_children, name=f"Child-{self._num_children}")
        self._num_children += 1
        print(f"Time: {Mundus.now}\tEvent Priority: {self.priority}:\t{self._name} picks up a child {c._name}.")

class child(Actor):

    def __init__(self, at, step, till, priority, name) -> None:
        super().__init__(at=at, step=step, till=till, priority=priority)
        self._name = name

    def action(self):
        if self.timeline.now >= 7:
            uncle.deactivate()
        print(f"Time: {Mundus.now}\tEvent Priority: {self.priority}:\t{self._name}\tis playing at home.")

m = mom(at=0, step=1, till=5, priority=0, name="Mom")
father = Actor(after=m, priority=-1, action=lambda: print(f"Time: {Mundus.now}\tEvent Priority: {father.priority}:\tFather comes back home."))
uncle = Actor(after=father, step=1, till=10, priority=0, action=lambda: print(f"Time: {Mundus.now}\tEvent Priority: {uncle.priority}:\tUncle comes to vist."))

Mundus.simulate(10)
print(m.status)
print(father.status)
print(uncle.status)



from __future__ import annotations
from copy import deepcopy
from typing import Generator, List, Optional, Union, Callable, TYPE_CHECKING
from uuid import uuid4
import warnings
from math import inf

from Akatosh import Timeline, Mundus

if TYPE_CHECKING:
    from Akatosh import Actor


class Consumer:
    _user: Actor
    _products: List

    def __init__(self, user: Actor) -> None:
        self._user = user
        self._products = list()

    def __eq__(self, __o: Consumer) -> bool:
        if self.user is __o.user:
            return True
        else:
            return False

    @property
    def user(self) -> Actor:
        return self._user

    @property
    def quantity(self) -> Union[int, float]:
        return len(self._products)

    @property
    def products(self) -> List:
        return self._products


class Producer:

    _id: int
    _action: Optional[Generator | Callable]
    _timeline: Timeline
    _priority: int
    _at: Union[int, float]
    _step: Union[int, float, Callable]
    _till: Union[int, float]
    _active: bool
    _after: Actor

    _time: Union[int, float]
    _status: List[str]
    _followers: List[Actor]

    _label: str
    _product: Callable
    _product_kargs: dict
    _production_rate: int
    _capacity: int
    _inventory: List
    _consumers: List[Consumer]

    def __init__(
        self,
        product: Callable,
        product_kargs: dict,
        prodction_rate: int,
        step: Union[int, float],
        timeline: Optional[Timeline] = None,
        priority: int = 0,
        at: Union[int, float, Callable] = 0,
        till: Optional[int | float | Callable] = None,
        active: bool = True,
        after: Optional[Actor] = None,
        label: Optional[str] = None,
        capacity: int = 1,
    ) -> None:

        # generate a unique id for this producer
        self._id = uuid4().int

        # assign the timeline to this producer
        if timeline is None:
            self._timeline = Mundus.timeline
        else:
            self._timeline = timeline
        self.timeline.actors.append(self)

        # assign the priority to this producer
        self._priority = priority

        # initialize the follower list
        self._followers = list()

        # assign the waiting target to this producer
        if after is not None:
            self._after = after
            after.followers.append(self)
            if after.priority > self.priority:
                warnings.warn(
                    message=f"Producer {self.id} has a lower priority than its waiting target {after.id}."
                )

        # initialize the time
        self._time = at
        self._till = till

        # initialize the status
        self._status = list()
        if active:
            self.status.append("active")
        else:
            self.status.append("inactive")
        if after is not None:
            self.status.append("onhold")

        # schedule the actor onto timeline
        if self.onhold is False:
            self.timeline.schedule(self)

        self._label = label or str()
        self._product = product
        self._product_kargs = product_kargs
        self._production_rate = prodction_rate
        self._step = step
        self._capacity = capacity
        self._inventory = list()
        self._consumers = list()

        if hasattr(self.product, "_priority"):
            if self.product.priority < self.priority:
                warnings.warn(
                    message=f"Producer {self} has lower priority than its producer's priority {self.priority}."
                )

    def perform(self):
        # non-continuous actor
        if self.step is None and self.till is None:
            if self.active:
                self.action()
        # continuous actor but misses step size
        elif self.step is None and self.till is not None:
            raise AttributeError(f"Actor {self.id} has no step size defined.")
        # continuous actor with step size but no end time
        elif self.step is not None and self.till is None:
            self._till = inf
            while self.time < self.till:
                if self.active:
                    self.action()
                if callable(self.step):
                    self._time += self.step()
                else:
                    self._time += self.step
                yield self.timeline.schedule(self)
            if self.active:
                self.action()
        # continuous actor with step size and end time
        elif self.step is not None and self.till is not None:
            if callable(self.till):
                self._till = self.till()
            while self.time < self.till:
                if self.active:
                    self.action()
                if callable(self.step):
                    self._time += self.step()
                else:
                    self._time += self.step
                yield self.timeline.schedule(self)
            if self.active:
                self.action()

    def action(self):
        for _ in range(self.production_rate):
            self.inventory.append(self.product(**self.product_kargs))
        print(f"Time {self.timeline.now}\t{self.label} has {len(self.inventory)} {self.product.__name__}.")

    def deactivate(self):
        if "active" in self.status:
            self.status.remove("active")
        if "inactive" not in self.status:
            self.status.append("inactive")

    def get(self, quantity: int = 1) -> List:
        if quantity <= len(self.inventory):
            return [self.inventory.pop() for _ in range(quantity)]
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available quantity {len(self.inventory)}."
            )

    def restock(self, quantity: int = 1) -> None:
        if len(self.inventory) + quantity <= self.capacity:
            self.inventory.extend([self.product for _ in range(quantity)])
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available capacity {self.capacity - len(self.inventory)}."
            )

    def distribute(self, user: Actor, quantity: Optional[int] = None) -> None:
        if quantity is None:
            consumer = Consumer(user)
            if consumer not in self.consumers:
                for _ in range(len(self.inventory)):
                    consumer.products.append(self.inventory.pop())
                self.consumers.append(consumer)
            else:
                for c in self.consumers:
                    if c.user is user:
                        for _ in range(len(self.inventory)):
                            consumer.products.append(self.inventory.pop())
        else:
            if isinstance(quantity, int):
                if quantity <= len(self.inventory):
                    consumer = Consumer(user)
                    if consumer not in self.consumers:
                        for _ in range(quantity):
                            consumer.products.append(self.inventory.pop())
                        self.consumers.append(consumer)
                    else:
                        for c in self.consumers:
                            if c.user is user:
                                for _ in range(quantity):
                                    consumer.products.append(self.inventory.pop())
                else:
                    raise ValueError(
                        f"Quantity {quantity} is greater than available quantity {len(self.inventory)}."
                    )
            else:
                raise TypeError(f"Quantity {quantity} is not an integer.")

    @property
    def label(self) -> str:
        return self._label

    @property
    def product(self) -> object:
        return self._product

    @property
    def product_kargs(self) -> dict:
        return self._product_kargs

    @property
    def production_rate(self) -> int:
        return self._production_rate

    @property
    def capacity(self) -> int:
        return self._capacity

    @property
    def inventory(self) -> List:
        return self._inventory

    @property
    def consumers(self) -> List[Consumer]:
        return [consumer.user for consumer in self._consumers]

    @property
    def id(self):
        return self._id

    @property
    def priority(self):
        return self._priority

    @property
    def timeline(self):
        return self._timeline

    @property
    def time(self):
        return self._time

    @property
    def till(self):
        return self._till

    @property
    def step(self):
        return self._step

    @property
    def status(self):
        return self._status

    @property
    def after(self):
        return self._after

    @property
    def active(self):
        return "active" in self.status

    @property
    def inactive(self):
        return "inactive" in self.status

    @property
    def onhold(self):
        return "onhold" in self.status

    @property
    def completed(self):
        return "completed" in self.status

    @property
    def followers(self):
        return self._followers



from __future__ import annotations
from typing import Union, List, Optional, TYPE_CHECKING
from dataclasses import dataclass
from uuid import uuid4

if TYPE_CHECKING:
    from Akatosh import Actor


@dataclass
class ResourceClaim:
    _user: Actor
    _quantity: Union[int, float]

    def __eq__(self, __o: ResourceClaim) -> bool:
        if self.user is __o.user:
            return True
        else:
            return False

    @property
    def user(self) -> Actor:
        return self._user

    @property
    def quantity(self) -> Union[int, float]:
        return self._quantity


class Resource:

    _id: int
    _label: str
    _capacity: Union[int, float]
    _claimed_quantity: Union[int, float]

    _claims: List[ResourceClaim]

    def __init__(
        self,
        label: Optional[str] = None,
        capacity: Union[int, float] = 1,
        init_claimed_quantity: Union[int, float] = 0,
    ) -> None:
        self._id = uuid4().int
        self._label = label or str()
        self._capacity = capacity
        self._claimed_quantity = init_claimed_quantity

        self._claims = list()

    def get(self, quantity: Union[int, float]) -> bool:
        if quantity <= self.available_quantity:
            self._claimed_quantity += quantity
            return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available quantity {self.available_quantity}."
            )

    def put(self, quantity: Union[int, float]) -> bool:
        if quantity <= self.claimed_quantity:
            self._claimed_quantity -= quantity
            return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than claimed quantity {self.claimed_quantity}."
            )

    def distribute(self, user: Actor, quantity: Union[int, float]) -> bool:
        if quantity <= self.available_quantity:
            self._claimed_quantity += quantity
            resource_claim = ResourceClaim(user, quantity)
            if resource_claim not in self.claims:
                self.claims.append(resource_claim)
                return True
            else:
                for claim in self.claims:
                    if claim.user is user:
                        claim._quantity += quantity
                        return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available quantity {self.available_quantity}."
            )

    def release(self, user: Optional[Actor | List[Actor]] = None, amount:Optional[int | float]=None) -> bool:
        if user is None:
            self._claimed_quantity = 0
            self._claims.clear()
            return True
        else:
            if isinstance(user, list):
                for u in user:
                    for claim in self.claims[:]:
                        if claim.user == u:
                            if amount is None:
                                if claim.quantity > self.claimed_quantity:
                                    raise ValueError(
                                        f"Resource claim's quantity {claim.quantity} is greater than resource's claimed quantity {self.claimed_quantity}."
                                    )
                                else:
                                    self._claimed_quantity -= claim.quantity
                                    self.claims.remove(claim)
                            else:
                                if amount > claim.quantity:
                                    raise ValueError(
                                        f"User tries to release {amount} units, but only {claim.quantity} units are claimed by the user."
                                    )
                                else:
                                    self._claimed_quantity -= amount
                                    claim._quantity -= amount
                                    if claim.quantity == 0:
                                        self.claims.remove(claim)
            else:
                for claim in self.claims[:]:
                    if claim.user == user:
                        if amount is None:
                            if claim.quantity > self.claimed_quantity:
                                raise ValueError(
                                    f"Resource claim's quantity {claim.quantity} is greater than resource's claimed quantity {self.claimed_quantity}."
                                )
                            else:
                                self._claimed_quantity -= claim.quantity
                                self.claims.remove(claim)
                        else:
                            if amount > claim.quantity:
                                raise ValueError(
                                    f"User tries to release {amount} units, but only {claim.quantity} units are claimed by the user."
                                )
                            else:
                                self._claimed_quantity -= amount
                                claim._quantity -= amount
                                if claim.quantity == 0:
                                    self.claims.remove(claim)

    @property
    def id(self) -> int:
        return self._id

    @property
    def label(self) -> str:
        return self._label

    @property
    def capacity(self) -> Union[int, float]:
        return self._capacity

    @property
    def claimed_quantity(self) -> Union[int, float]:
        return self._claimed_quantity

    @property
    def available_quantity(self) -> Union[int, float]:
        return self.capacity - self.claimed_quantity

    @property
    def claims(self) -> List[ResourceClaim]:
        return self._claims

    @property
    def users(self):
        return [claim.user for claim in self.claims]



from Akatosh import Actor, Producer, Mundus


class Milk(Actor):
    def __init__(
        self,
        expiration_time: int,
        **kwargs,
    ) -> None:
        super().__init__(step=1, priority=1, **kwargs)
        self._expiration_time = expiration_time
        self._life_time = 0
        self._expired = False

    def action(self):
        self._life_time += 1
        if self.expired is False:
            if self._life_time > self._expiration_time:
                self._expired = True
                print(f"Time {self.timeline.now}\tMilk {self.id} has expired.")

    @property
    def expired(self) -> bool:
        return self._expired

    @property
    def expiration_time(self) -> int:
        return self._expiration_time


class MilkFarm(Producer):
    def __init__(
        self,
        production_cycle: int = 1,
        num_of_bottles: int = 1,
        **kwargs,
    ) -> None:
        super().__init__(
            product=Milk,
            prodction_rate=num_of_bottles,
            step=production_cycle,
            product_kargs={"expiration_time": 3},
            priority=0,
            label="Milk Farm",
            **kwargs,
        )


class Client(Actor):
    def __init__(
        self,
        restock_period: int = 1,
        restock_quantity: int = 50,
        farm: MilkFarm = None,
        **kwargs,
    ) -> None:
        super().__init__(step=restock_period, priority=2, **kwargs)
        self._restock_quantity = restock_quantity
        self._farm = farm

    def action(self):
        stock = 0
        while stock < self.restock_quantity:
            try:
                self.consume(self._farm, 1)
                stock += 1
            except Exception as e:
                break
        print(
            f"Time {self.timeline.now}\tClient {self.id} has restocked {stock} bottles of milk."
        )

    @property
    def restock_quantity(self) -> int:
        return self._restock_quantity


farm = MilkFarm(1, 10)
client = Client(3, 50, farm)


Mundus.simulate(20)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union
from uuid import uuid4
import inspect

from Akatosh import Event

if TYPE_CHECKING:
    from Akatosh import Actor


class Timeline:
    _id: int
    _time: Union[int, float]
    _actors: List[Actor]
    _events: List[Event]

    def __init__(self) -> None:
        self._id = uuid4().int
        self._time = 0
        self._actors = list()
        self._events = list()

    def schedule(self, actor: Actor):
        event = Event(at=actor.time, priority=actor.priority, actor=actor)
        self.events.append(event)
        self.events.sort(key=lambda event: event.priority)
        self.events.sort(key=lambda event: event.at)
        if actor.status.count('scheduled') == 0:
            actor.status.append('scheduled')

    def forward(self, till: Union[int, float]):
        while True:
            if len(self.events) != 0:
                next_event = self.events.pop(0)
                if self.now < next_event.at:
                    self._time = next_event.at
                if next_event.at <= till:
                    if inspect.isgeneratorfunction(next_event.actor.perform):
                        try:
                            next(next_event.actor.perform())
                        except StopIteration:
                            next_event.actor.status.append('completed')
                            for actor in next_event.actor.followers:
                                actor.status.remove('onhold')
                                actor._time = next_event.actor.time
                                self.schedule(actor)
                    else:
                        try:
                            next_event.actor.perform()
                            next_event.actor.status.append('completed')
                            for actor in next_event.actor.followers:
                                actor.status.remove('onhold')
                                actor._time = self.now
                                self.schedule(actor)
                        except Exception as e:
                            print(e)
                else:
                    break
            else:
                break
            
    def print_events(self):
        for event in self.events:
            print(f"{event.at}\t{event.priority}\t{event.actor}\t{event.actor.status}")
            
    @property
    def now(self):
        return self._time

    @property
    def actors(self):
        return self._actors

    @property
    def events(self):
        return self._events



from __future__ import annotations
from tkinter import W
from typing import Generator, List, Optional, Union, Callable, TYPE_CHECKING
from uuid import uuid4
from math import inf
import warnings

from Akatosh import Timeline, Mundus

if TYPE_CHECKING:
    from Akatosh import Resource
    from Akatosh import Producer


class Actor:

    _id: int
    _action: Optional[Generator | Callable]
    _timeline: Timeline
    _priority: int
    _at: Union[int, float]
    _step: Union[int, float, Callable]
    _till: Union[int, float]
    _active: bool
    _after: Actor

    _time: Union[int, float]
    _status: List[str]
    _followers: List[Actor]

    def __init__(
        self,
        action: Optional[Generator | Callable] = None,
        timeline: Optional[Timeline] = None,
        priority: int = 0,
        at: Union[int, float, Callable] = 0,
        step: Optional[int | float | Callable] = None,
        till: Optional[int | float | Callable] = None,
        active: bool = True,
        after: Optional[Actor] = None,
    ) -> None:

        # generate a unique id for this actor
        self._id = uuid4().int

        # assign the action to this actor
        self._action = action

        # assign the timeline to this actor
        if timeline is None:
            self._timeline = Mundus.timeline
        else:
            self._timeline = timeline
        self.timeline.actors.append(self)

        # assign the priority to this actor
        self._priority = priority

        # initialize the follower list
        self._followers = list()

        # assign the waiting target to this actor
        if after is not None:
            self._after = after
            after.followers.append(self)
            if after.priority > self.priority:
                warnings.warn(
                    message = f"Actor {self.id} has a lower priority than its waiting target {after.id}."
                )

        # initialize the time
        self._time = at
        self._step = step
        self._till = till

        # initialize the status
        self._status = list()
        if active:
            self.status.append("active")
        else:
            self.status.append("inactive")
        if after is not None:
            self.status.append("onhold")

        # schedule the actor onto timeline
        if self.onhold is False:
            self.timeline.schedule(self)

    def perform(self):
        # non-continuous actor
        if self.step is None and self.till is None:
            if self.active:
                self.action()
        # continuous actor but misses step size
        elif self.step is None and self.till is not None:
            raise AttributeError(f"Actor {self.id} has no step size defined.")
        # continuous actor with step size but no end time
        elif self.step is not None and self.till is None:
            self._till = inf
            while self.time < self.till:
                if self.active:
                    self.action()
                if callable(self.step):
                    self._time += self.step()
                else:
                    self._time += self.step
                yield self.timeline.schedule(self)
            if self.active:
                self.action()
        # continuous actor with step size and end time
        elif self.step is not None and self.till is not None:
            if callable(self.till):
                self._till = self.till()
            while self.time < self.till:
                if self.active:
                    self.action()
                if callable(self.step):
                    self._time += self.step()
                else:
                    self._time += self.step
                yield self.timeline.schedule(self)
            if self.active:
                self.action()

    def deactivate(self):
        if "active" in self.status:
            self.status.remove("active")
        if "inactive" not in self.status:
            self.status.append("inactive")

    def request(self, resource: Resource, amount: Union[int, float] = 1) -> bool:
        resource.distribute(self, amount)

    def release(self, resource: Resource, amount: Optional[int | float] = None) -> bool:
        if amount is None:
            resource.release(self)
        else:
            resource.release(self, amount)

    def consume(self, producer: Producer, amount: Optional[int | float] = None) -> bool:
        producer.distribute(self, amount)

    @property
    def id(self):
        return self._id

    @property
    def priority(self):
        return self._priority

    @property
    def action(self):
        return self._action

    @property
    def timeline(self):
        return self._timeline

    @property
    def time(self):
        return self._time

    @property
    def till(self):
        return self._till

    @property
    def step(self):
        return self._step

    @property
    def status(self):
        return self._status

    @property
    def after(self):
        return self._after

    @property
    def active(self):
        return "active" in self.status

    @property
    def inactive(self):
        return "inactive" in self.status

    @property
    def onhold(self):
        return "onhold" in self.status

    @property
    def completed(self):
        return "completed" in self.status

    @property
    def followers(self):
        return self._followers



from ctypes import Union
from math import inf
from typing import List, Union
from uuid import uuid4

from .timeline import Timeline


class Universe:
    _id: int
    _timeline: Timeline
    _till: Union[int, float]

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(Universe, cls).__new__(cls)
        return cls.instance

    def __init__(self) -> None:
        self._id = uuid4().int
        self._timeline = Timeline()
        self._till = int()

    def simulate(self, till: Union[int, float] = inf) -> None:
        self._till = till
        while self.timeline.now <= self.till:
            self.timeline.forward(self.till)
            if len(self.timeline.events) == 0:
                break

    @property
    def timeline(self) -> Timeline:
        return self._timeline

    @property
    def till(self) -> Union[int, float]:
        return self._till

    @property
    def now(self) -> Union[int, float]:
        return self.timeline.now


Mundus = Universe()
